# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import shutil
import tarfile
from collections import OrderedDict, defaultdict
from multiprocessing.pool import ThreadPool

from pex import hashing, resolver
from pex.auth import PasswordDatabase
from pex.build_system import pep_517
from pex.common import open_zip, pluralize, safe_mkdtemp
from pex.dist_metadata import DistMetadata, ProjectNameAndVersion
from pex.fetcher import URLFetcher
from pex.jobs import Job, Retain, SpawnedJob, execute_parallel
from pex.orderedset import OrderedSet
from pex.pep_503 import ProjectName
from pex.pip.download_observer import DownloadObserver
from pex.pip.tool import PackageIndexConfiguration
from pex.resolve import lock_resolver, locker, resolvers
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.downloads import ArtifactDownloader
from pex.resolve.locked_resolve import (
    Artifact,
    FileArtifact,
    LocalProjectArtifact,
    LockConfiguration,
    LockedResolve,
    LockStyle,
    VCSArtifact,
)
from pex.resolve.locker import Locker
from pex.resolve.lockfile.download_manager import DownloadManager
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.pep_691.fingerprint_service import FingerprintService
from pex.resolve.requirement_configuration import RequirementConfiguration
from pex.resolve.resolved_requirement import Pin, ResolvedRequirement
from pex.resolve.resolver_configuration import PipConfiguration
from pex.resolve.resolvers import Resolver
from pex.resolver import BuildRequest, Downloaded, ResolveObserver, WheelBuilder
from pex.result import Error, try_
from pex.targets import Target, Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING
from pex.version import __version__

if TYPE_CHECKING:
    from typing import DefaultDict, Dict, Iterable, List, Mapping, Optional, Tuple, Union

    import attr  # vendor:skip

    from pex.hashing import HintedDigest
    from pex.requirements import ParsedRequirement

    AnyArtifact = Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
else:
    from pex.third_party import attr


class CreateLockDownloadManager(DownloadManager[Artifact]):
    @classmethod
    def create(
        cls,
        download_dir,  # type: str
        locked_resolves,  # type: Iterable[LockedResolve]
        pex_root=None,  # type: Optional[str]
    ):
        # type: (...) -> CreateLockDownloadManager

        file_artifacts_by_filename = {}  # type: Dict[str, Tuple[FileArtifact, ProjectName]]
        source_artifacts_by_pin = (
            {}
        )  # type: Dict[Pin, Tuple[Union[LocalProjectArtifact, VCSArtifact], ProjectName]]
        for locked_resolve in locked_resolves:
            for locked_requirement in locked_resolve.locked_requirements:
                pin = locked_requirement.pin
                project_name = pin.project_name
                for artifact in locked_requirement.iter_artifacts():
                    if isinstance(artifact, FileArtifact):
                        file_artifacts_by_filename[artifact.filename] = (artifact, project_name)
                    else:
                        # N.B.: We know there is only ever one local project artifact for a given
                        # locked local project requirement and likewise only one VCS artifact for a
                        # given locked VCS requirement.
                        source_artifacts_by_pin[locked_requirement.pin] = (artifact, project_name)

        path_by_artifact_and_project_name = {}  # type: Dict[Tuple[Artifact, ProjectName], str]
        for root, _, files in os.walk(download_dir):
            for f in files:
                artifact_and_project_name = file_artifacts_by_filename.get(
                    f
                )  # type: Optional[Tuple[AnyArtifact, ProjectName]]
                if not artifact_and_project_name:
                    project_name_and_version = ProjectNameAndVersion.from_filename(f)
                    pin = Pin.canonicalize(project_name_and_version)
                    artifact_and_project_name = source_artifacts_by_pin[pin]
                path_by_artifact_and_project_name[artifact_and_project_name] = os.path.join(root, f)

        return cls(
            path_by_artifact_and_project_name=path_by_artifact_and_project_name, pex_root=pex_root
        )

    def __init__(
        self,
        path_by_artifact_and_project_name,  # type: Mapping[Tuple[Artifact, ProjectName], str]
        pex_root=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        super(CreateLockDownloadManager, self).__init__(pex_root=pex_root)
        self._path_by_artifact_and_project_name = path_by_artifact_and_project_name

    def store_all(self):
        # type: () -> None
        for artifact, project_name in self._path_by_artifact_and_project_name:
            self.store(artifact, project_name)

    def save(
        self,
        artifact,  # type: Artifact
        project_name,  # type: ProjectName
        dest_dir,  # type: str
        digest,  # type: HintedDigest
    ):
        # type: (...) -> Union[str, Error]
        src = self._path_by_artifact_and_project_name[(artifact, project_name)]
        filename = os.path.basename(src)
        dest = os.path.join(dest_dir, filename)
        shutil.move(src, dest)

        hashing.file_hash(dest, digest=digest)
        return filename


@attr.s(frozen=True)
class _LockAnalysis(object):
    target = attr.ib()  # type: Target
    analyzer = attr.ib()  # type: Locker
    download_dir = attr.ib()  # type: str


class LockError(Exception):
    """Indicates an error creating a lock file."""


def _prepare_project_directory(build_request):
    # type: (BuildRequest) -> Tuple[Target, str]

    project = build_request.source_path
    target = build_request.target
    if os.path.isdir(project):
        return target, project

    extract_dir = os.path.join(safe_mkdtemp(), "project")
    if project.endswith(".zip"):
        with open_zip(project) as zf:
            zf.extractall(extract_dir)
    elif project.endswith(".tar.gz"):
        with tarfile.open(project) as tf:
            tf.extractall(extract_dir)
    else:
        raise LockError("Unexpected archive type for sdist {project}".format(project=project))

    listing = os.listdir(extract_dir)
    if len(listing) != 1:
        raise LockError(
            "Expected one top-level project directory to be extracted from {project}, "
            "found {count}: {listing}".format(
                project=project, count=len(listing), listing=", ".join(listing)
            )
        )
    return target, os.path.join(extract_dir, listing[0])


@attr.s(frozen=True)
class LockObserver(ResolveObserver):
    root_requirements = attr.ib()  # type: Tuple[ParsedRequirement, ...]
    lock_configuration = attr.ib()  # type: LockConfiguration
    resolver = attr.ib()  # type: Resolver
    wheel_builder = attr.ib()  # type: WheelBuilder
    package_index_configuration = attr.ib()  # type: PackageIndexConfiguration
    max_parallel_jobs = attr.ib(default=None)  # type: Optional[int]
    _analysis = attr.ib(factory=OrderedSet, eq=False)  # type: OrderedSet[_LockAnalysis]

    def observe_download(
        self,
        target,  # type: Target
        download_dir,  # type: str
    ):
        # type: (...) -> DownloadObserver
        analyzer = Locker(
            target=target,
            root_requirements=self.root_requirements,
            pip_version=self.package_index_configuration.pip_version,
            resolver=self.resolver,
            lock_configuration=self.lock_configuration,
            download_dir=download_dir,
            fingerprint_service=FingerprintService.create(
                url_fetcher=URLFetcher(
                    network_configuration=self.package_index_configuration.network_configuration,
                    password_entries=self.package_index_configuration.password_entries,
                ),
                max_parallel_jobs=self.max_parallel_jobs,
            ),
        )
        patch_set = locker.patch(lock_configuration=self.lock_configuration)
        observer = DownloadObserver(analyzer=analyzer, patch_set=patch_set)
        self._analysis.add(
            _LockAnalysis(target=target, analyzer=analyzer, download_dir=download_dir)
        )
        return observer

    def _spawn_prepare_metadata(self, target_and_project_directory):
        # type: (Tuple[Target, str]) -> SpawnedJob[DistMetadata]

        target, project_directory = target_and_project_directory
        return pep_517.spawn_prepare_metadata(
            project_directory=project_directory,
            pip_version=self.package_index_configuration.pip_version,
            target=target,
            resolver=self.resolver,
        )

    def lock(self, downloaded):
        # type: (Downloaded) -> Tuple[LockedResolve, ...]

        dist_metadatas_by_target = defaultdict(
            OrderedSet
        )  # type: DefaultDict[Target, OrderedSet[DistMetadata]]
        build_requests = OrderedSet()  # type: OrderedSet[BuildRequest]

        for local_distribution in downloaded.local_distributions:
            if local_distribution.is_wheel:
                dist_metadatas_by_target[local_distribution.target].add(
                    DistMetadata.load(local_distribution.path)
                )
            else:
                build_requests.add(
                    BuildRequest.create(
                        target=local_distribution.target, source_path=local_distribution.path
                    )
                )

        resolved_requirements_by_target = (
            OrderedDict()
        )  # type: OrderedDict[Target, Tuple[ResolvedRequirement, ...]]
        for analysis in self._analysis:
            lock_result = analysis.analyzer.lock_result
            build_requests.update(
                BuildRequest.create(target=analysis.target, source_path=local_project)
                for local_project in lock_result.local_projects
            )
            resolved_requirements_by_target[analysis.target] = lock_result.resolved_requirements

        with TRACER.timed(
            "Building {count} source {distributions} to gather metadata for lock.".format(
                count=len(build_requests), distributions=pluralize(build_requests, "distribution")
            )
        ):
            pool = ThreadPool(processes=self.max_parallel_jobs)
            try:
                targets_and_project_directories = list(
                    pool.map(_prepare_project_directory, build_requests),
                )
            finally:
                pool.close()
                pool.join()

            build_wheel_requests = []  # type: List[BuildRequest]
            prepare_metadata_errors = OrderedDict()  # type: OrderedDict[str, str]
            for build_request, dist_metadata_result in zip(
                build_requests,
                # MyPy can't infer the _I type argument of Tuple[Target, str] here.
                execute_parallel(  # type: ignore[misc]
                    targets_and_project_directories,
                    # MyPy just can't figure out the next two args types; they're OK.
                    self._spawn_prepare_metadata,  # type: ignore[arg-type]
                    error_handler=Retain[str](),  # type: ignore[arg-type]
                    max_jobs=self.max_parallel_jobs,
                ),
            ):
                if isinstance(dist_metadata_result, DistMetadata):
                    dist_metadatas_by_target[build_request.target].add(dist_metadata_result)
                else:
                    _item, error = dist_metadata_result
                    if isinstance(error, Job.Error) and pep_517.is_hook_unavailable_error(error):
                        TRACER.log(
                            "Failed to prepare metadata for {project}, trying to build a wheel "
                            "instead: {err}".format(
                                project=build_request.source_path, err=dist_metadata_result
                            ),
                            V=3,
                        )
                        build_wheel_requests.append(build_request)
                    else:
                        prepare_metadata_errors[build_request.source_path] = str(error)

            if prepare_metadata_errors:
                raise LockError(
                    "Could not gather lock metadata for {count} {projects} with source artifacts:\n"
                    "{errors}".format(
                        count=len(prepare_metadata_errors),
                        projects=pluralize(prepare_metadata_errors, "project"),
                        errors="\n".join(
                            "{index}. {project}: {error}".format(
                                index=index, project=project, error=error
                            )
                            for index, (project, error) in enumerate(
                                prepare_metadata_errors.items(), start=1
                            )
                        ),
                    )
                )

            if build_wheel_requests:
                build_wheel_results = self.wheel_builder.build_wheels(
                    build_requests=build_wheel_requests,
                    max_parallel_jobs=self.max_parallel_jobs,
                )
                for install_requests in build_wheel_results.values():
                    for install_request in install_requests:
                        dist_metadatas_by_target[install_request.target].add(
                            DistMetadata.load(install_request.wheel_path)
                        )

        return tuple(
            LockedResolve.create(
                resolved_requirements=resolved_requirements,
                dist_metadatas=dist_metadatas_by_target[target],
                fingerprinter=ArtifactDownloader(
                    resolver=self.resolver,
                    target=target,
                    package_index_configuration=self.package_index_configuration,
                    max_parallel_jobs=self.max_parallel_jobs,
                ),
                platform_tag=None
                if self.lock_configuration.style == LockStyle.UNIVERSAL
                else target.platform.tag,
            )
            for target, resolved_requirements in resolved_requirements_by_target.items()
        )


def create(
    lock_configuration,  # type: LockConfiguration
    requirement_configuration,  # type: RequirementConfiguration
    targets,  # type: Targets
    pip_configuration,  # type: PipConfiguration
):
    # type: (...) -> Union[Lockfile, Error]
    """Create a lock file for the given resolve configurations."""

    network_configuration = pip_configuration.network_configuration
    parsed_requirements = tuple(requirement_configuration.parse_requirements(network_configuration))
    constraints = tuple(
        parsed_constraint.requirement
        for parsed_constraint in requirement_configuration.parse_constraints(network_configuration)
    )

    package_index_configuration = PackageIndexConfiguration.create(
        pip_version=pip_configuration.version,
        resolver_version=pip_configuration.resolver_version,
        network_configuration=network_configuration,
        find_links=pip_configuration.repos_configuration.find_links,
        indexes=pip_configuration.repos_configuration.indexes,
        password_entries=PasswordDatabase.from_netrc()
        .append(pip_configuration.repos_configuration.password_entries)
        .entries,
    )

    configured_resolver = ConfiguredResolver(pip_configuration=pip_configuration)
    lock_observer = LockObserver(
        root_requirements=parsed_requirements,
        lock_configuration=lock_configuration,
        resolver=configured_resolver,
        wheel_builder=WheelBuilder(
            package_index_configuration=package_index_configuration,
            prefer_older_binary=pip_configuration.prefer_older_binary,
            use_pep517=pip_configuration.use_pep517,
            build_isolation=pip_configuration.build_isolation,
            pip_version=pip_configuration.version,
            resolver=configured_resolver,
        ),
        package_index_configuration=package_index_configuration,
        max_parallel_jobs=pip_configuration.max_jobs,
    )

    download_dir = safe_mkdtemp()

    if lock_configuration.style is LockStyle.UNIVERSAL:
        download_targets = (
            Targets(interpreters=(targets.interpreter,)) if targets.interpreter else Targets()
        )
    else:
        download_targets = targets

    try:
        downloaded = resolver.download(
            targets=download_targets,
            requirements=requirement_configuration.requirements,
            requirement_files=requirement_configuration.requirement_files,
            constraint_files=requirement_configuration.constraint_files,
            allow_prereleases=pip_configuration.allow_prereleases,
            transitive=pip_configuration.transitive,
            indexes=pip_configuration.repos_configuration.indexes,
            find_links=pip_configuration.repos_configuration.find_links,
            resolver_version=pip_configuration.resolver_version,
            network_configuration=network_configuration,
            password_entries=pip_configuration.repos_configuration.password_entries,
            build=pip_configuration.allow_builds,
            use_wheel=pip_configuration.allow_wheels,
            prefer_older_binary=pip_configuration.prefer_older_binary,
            use_pep517=pip_configuration.use_pep517,
            build_isolation=pip_configuration.build_isolation,
            max_parallel_jobs=pip_configuration.max_jobs,
            observer=lock_observer,
            dest=download_dir,
            preserve_log=pip_configuration.preserve_log,
            pip_version=pip_configuration.version,
            resolver=configured_resolver,
        )
    except resolvers.ResolveError as e:
        return Error(str(e))

    with TRACER.timed("Creating lock from resolve"):
        locked_resolves = lock_observer.lock(downloaded)

    with TRACER.timed("Indexing downloads"):
        create_lock_download_manager = CreateLockDownloadManager.create(
            download_dir=download_dir, locked_resolves=locked_resolves
        )
        create_lock_download_manager.store_all()

    lock = Lockfile.create(
        pex_version=__version__,
        style=lock_configuration.style,
        requires_python=lock_configuration.requires_python,
        target_systems=lock_configuration.target_systems,
        pip_version=pip_configuration.version,
        resolver_version=pip_configuration.resolver_version,
        requirements=parsed_requirements,
        constraints=constraints,
        allow_prereleases=pip_configuration.allow_prereleases,
        allow_wheels=pip_configuration.allow_wheels,
        allow_builds=pip_configuration.allow_builds,
        prefer_older_binary=pip_configuration.prefer_older_binary,
        use_pep517=pip_configuration.use_pep517,
        build_isolation=pip_configuration.build_isolation,
        transitive=pip_configuration.transitive,
        locked_resolves=locked_resolves,
    )

    if lock_configuration.style is LockStyle.UNIVERSAL and (
        targets.platforms or targets.complete_platforms
    ):
        check_targets = Targets(
            platforms=targets.platforms,
            complete_platforms=targets.complete_platforms,
            assume_manylinux=targets.assume_manylinux,
        )
        with TRACER.timed(
            "Checking lock can resolve for platforms: {targets}".format(targets=check_targets)
        ):
            try_(
                lock_resolver.resolve_from_lock(
                    targets=check_targets,
                    lock=lock,
                    resolver=configured_resolver,
                    indexes=pip_configuration.repos_configuration.indexes,
                    find_links=pip_configuration.repos_configuration.find_links,
                    resolver_version=pip_configuration.resolver_version,
                    network_configuration=network_configuration,
                    password_entries=pip_configuration.repos_configuration.password_entries,
                    build=pip_configuration.allow_builds,
                    use_wheel=pip_configuration.allow_wheels,
                    prefer_older_binary=pip_configuration.prefer_older_binary,
                    use_pep517=pip_configuration.use_pep517,
                    build_isolation=pip_configuration.build_isolation,
                    transitive=pip_configuration.transitive,
                    max_parallel_jobs=pip_configuration.max_jobs,
                    pip_version=pip_configuration.version,
                )
            )

    return lock



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json
import os

# N.B.: The following environment variable is used by the Pex runtime to control Pip and must be
# kept in-sync with `__init__.py`.
with open(os.environ.pop("_PEX_PYTHON_VERSIONS_FILE")) as fp:
    PYTHON_FULL_VERSIONS = json.load(fp)
PYTHON_VERSIONS = sorted(set((version[0], version[1]) for version in PYTHON_FULL_VERSIONS))


def patch():
    # The pip-legacy-resolver patch.
    from pip._internal.utils import packaging  # type: ignore[import]

    if PYTHON_FULL_VERSIONS:
        orig_check_requires_python = packaging.check_requires_python

        def check_requires_python(requires_python, *_args, **_kw):
            # Ensure any dependency we lock is compatible with the full interpreter range
            # specified since we have no way to force Pip to backtrack and follow paths for any
            # divergences. Most (all?) true divergences should be covered by forked environment
            # markers.
            return all(
                orig_check_requires_python(requires_python, python_full_version)
                for python_full_version in PYTHON_FULL_VERSIONS
            )

        packaging.check_requires_python = check_requires_python
    else:
        packaging.check_requires_python = lambda *_args, **_kw: True

    # The pip-2020-resolver patch.
    from pip._internal.resolution.resolvelib.candidates import (  # type: ignore[import]
        RequiresPythonCandidate,
    )
    from pip._internal.resolution.resolvelib.requirements import (  # type: ignore[import]
        RequiresPythonRequirement,
    )

    if PYTHON_FULL_VERSIONS:
        orig_get_candidate_lookup = RequiresPythonRequirement.get_candidate_lookup
        orig_is_satisfied_by = RequiresPythonRequirement.is_satisfied_by

        # Ensure we do a proper, but minimal, comparison for Python versions. Previously we
        # always tested all `Requires-Python` specifier sets against Python full versions.
        # That can be pathologically slow (see:
        # https://github.com/pantsbuild/pants/issues/14998); so we avoid using Python full
        # versions unless the `Requires-Python` specifier set requires that data. In other
        # words:
        #
        # Need full versions to evaluate properly:
        # + Requires-Python: >=3.7.6
        # + Requires-Python: >=3.7,!=3.7.6,<4
        #
        # Do not need full versions to evaluate properly:
        # + Requires-Python: >=3.7,<4
        # + Requires-Python: ==3.7.*
        # + Requires-Python: >=3.6.0
        #
        def needs_full_versions(spec):
            components = spec.version.split(".", 2)
            if len(components) < 3:
                return False
            major_, minor_, patch = components
            if spec.operator in ("<", "<=", ">", ">=") and patch == "0":
                return False
            return patch != "*"

        def _py_versions(self):
            if not hasattr(self, "__py_versions"):
                self.__py_versions = (
                    version
                    for version in (
                        PYTHON_FULL_VERSIONS
                        if any(needs_full_versions(spec) for spec in self.specifier)
                        else PYTHON_VERSIONS
                    )
                    if ".".join(map(str, version)) in self.specifier
                )
            return self.__py_versions

        def get_candidate_lookup(self):
            for py_version in self._py_versions():
                delegate = RequiresPythonRequirement(
                    self.specifier, RequiresPythonCandidate(py_version)
                )
                candidate_lookup = orig_get_candidate_lookup(delegate)
                if candidate_lookup != (None, None):
                    return candidate_lookup
            return None, None

        def is_satisfied_by(self, *_args, **_kw):
            # Ensure any dependency we lock is compatible with the full interpreter range
            # specified since we have no way to force Pip to backtrack and follow paths for any
            # divergences. Most (all?) true divergences should be covered by forked environment
            # markers.
            return all(
                orig_is_satisfied_by(self, RequiresPythonCandidate(py_version))
                for py_version in self._py_versions()
            )

        RequiresPythonRequirement._py_versions = _py_versions
        RequiresPythonRequirement.get_candidate_lookup = get_candidate_lookup
        RequiresPythonRequirement.is_satisfied_by = is_satisfied_by
    else:
        RequiresPythonRequirement.get_candidate_lookup = lambda self: (self._candidate, None)
        RequiresPythonRequirement.is_satisfied_by = lambda *_args, **_kw: True



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.compatibility import to_bytes
from pex.executor import Executor
from pex.interpreter import PythonInterpreter
from pex.typing import TYPE_CHECKING, cast
from pex.util import named_temporary_file

if TYPE_CHECKING:
    from typing import Iterable, List, Text


_COMPILER_MAIN = """
from __future__ import print_function

import os
import py_compile
import sys


def compile(root, relpaths):
  compiled = []
  errored = {}
  for relpath in relpaths:
    abspath = os.path.join(root, relpath)
    # NB: We give the compiled bytecode file a `.pyc` extension, but if PYTHONOPTIMIZE is in play
    # the generated bytecode will be optimized.  Traditionally these optimized bytecode files would
    # have a `.pyo` extension, but the extension only matters for location of the file to execute
    # for a given module and not on the interpretation of its bytecode contents.  As such we're
    # safe to pick the `.pyc` extension for all bytecode file cases without a need to interpret the
    # current optimization setting for the active python interpreter.
    pyc_relpath = relpath + 'c'
    pyc_abspath = os.path.join(root, pyc_relpath)
    try:
      py_compile.compile(abspath, cfile=pyc_abspath, dfile=relpath, doraise=True)
      compiled.append(pyc_relpath)
    except py_compile.PyCompileError as e:
      errored[e.file] = e.msg
  return compiled, errored


def main(root, relpaths):
  compiled, errored = compile(root, relpaths)
  if not errored:
    for path in compiled:
      print(path)
    sys.exit(0)

  print('Encountered %%d errors compiling %%d files:' %% (len(errored), len(relpaths)),
        file=sys.stderr)
  for file, msg in errored.items():
    print('  %%s: %%s' %% (file, msg), file=sys.stderr)
  sys.exit(1)

root = %(root)r
relpaths = %(relpaths)r

main(root, relpaths)
"""


class Compiler(object):
    class Error(Exception):
        pass

    # N.B. This subclasses `Error` only for backwards compatibility.
    class CompilationFailure(Error):
        """Indicates an error compiling one or more python source files."""

    def __init__(self, interpreter):
        # type: (PythonInterpreter) -> None
        """Creates a bytecode compiler for the given `interpreter`.

        :param interpreter: The interpreter to use to compile sources with.
        """
        self._interpreter = interpreter

    def compile(self, root, relpaths):
        # type: (str, Iterable[str]) -> List[Text]
        """Compiles the given python source files using this compiler's interpreter.

        :param root: The root path all the source files are found under.
        :param relpaths: The relative paths from the `root` of the source files to compile.
        :returns: A list of relative paths of the compiled bytecode files.
        :raises: A :class:`Compiler.Error` if there was a problem bytecode compiling any of the files.
        """
        with named_temporary_file() as fp:
            fp.write(
                to_bytes(_COMPILER_MAIN % {"root": root, "relpaths": relpaths}, encoding="utf-8")
            )
            fp.flush()

            try:
                _, out, _ = self._interpreter.execute(args=[fp.name])
            except Executor.NonZeroExit as e:
                raise self.CompilationFailure(
                    "encountered %r during bytecode compilation.\nstderr was:\n%s\n" % (e, e.stderr)
                )

            return cast("Text", out).splitlines()



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import os
import shutil
import subprocess
import sys

from pex.pex_info import PexInfo
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING
from pex.variables import unzip_dir

if TYPE_CHECKING:
    from typing import Any


def test_layout_identification(tmpdir):
    # type: (Any) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    pex_file = os.path.join(str(tmpdir), "a.pex")
    run_pex_command(
        args=["-o", pex_file, "--pex-root", pex_root, "--runtime-pex-root", pex_root]
    ).assert_success()

    pex_hash = PexInfo.from_pex(pex_file).pex_hash
    assert pex_hash is not None

    expected_unzip_dir = unzip_dir(pex_root, pex_hash)
    assert not os.path.exists(expected_unzip_dir)

    subprocess.check_call(args=[pex_file, "-c", ""])
    assert os.path.isdir(expected_unzip_dir)

    shutil.rmtree(expected_unzip_dir)
    os.chmod(pex_file, 0o644)
    subprocess.check_call(args=[sys.executable, pex_file, "-c", ""])
    assert os.path.isdir(expected_unzip_dir)



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import errno
import multiprocessing
import os
import shutil
import subprocess
import sys
import tempfile
from subprocess import CalledProcessError
from textwrap import dedent

import pytest

from pex.common import safe_mkdtemp, safe_open, temporary_dir, touch
from pex.compatibility import PY2
from pex.executor import Executor
from pex.interpreter import PythonInterpreter
from pex.layout import Layout
from pex.pex_builder import CopyMode, PEXBuilder
from pex.testing import IS_PYPY, PY310, PY_VER, ensure_python_interpreter, run_pex_command
from pex.typing import TYPE_CHECKING, cast
from pex.util import named_temporary_file
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, Iterator, List, Optional, Protocol, Set, Text, Tuple

    class CreatePexVenv(Protocol):
        def __call__(self, *options):
            # type: (*str) -> Virtualenv
            pass


FABRIC_VERSION = "2.5.0"


@pytest.fixture(scope="module")
def pex():
    # type: () -> Iterator[str]
    with temporary_dir() as tmpdir:
        pex_path = os.path.join(tmpdir, "fabric.pex")

        src_dir = os.path.join(tmpdir, "src")
        touch(os.path.join(src_dir, "user/__init__.py"))
        touch(os.path.join(src_dir, "user/package/__init__.py"))

        # Fabric dips into Invoke vendored code. It depends on "invoke<2.0,>=1.3", but in version
        # 1.7.0, the vendored `decorator` module Fabric depends on inside Invoke no longer is
        # importable under Python 2.7; so we pin low.
        constraints = os.path.join(tmpdir, "constraints.txt")
        with open(constraints, "w") as fp:
            fp.write("Invoke==1.6.0")

        # N.B.: --unzip just speeds up runs 2+ of the pex file and is otherwise not relevant to
        # these tests.
        run_pex_command(
            args=[
                "fabric=={}".format(FABRIC_VERSION),
                "--constraints",
                constraints,
                "-c",
                "fab",
                "--sources-directory",
                src_dir,
                "-o",
                pex_path,
                "--include-tools",
            ]
        )
        yield os.path.realpath(pex_path)


def make_env(**kwargs):
    # type: (**Any) -> Dict[str, str]
    env = os.environ.copy()
    env.update((k, str(v)) for k, v in kwargs.items())
    return env


@pytest.fixture
def create_pex_venv(pex):
    # type: (str) -> Iterator[CreatePexVenv]
    with temporary_dir() as tmpdir:
        venv_dir = os.path.join(tmpdir, "venv")

        def _create_pex_venv(*options):
            # type: (*str) -> Virtualenv
            subprocess.check_call(
                args=[pex, "venv", venv_dir] + list(options or ()), env=make_env(PEX_TOOLS="1")
            )
            return Virtualenv(venv_dir)

        yield _create_pex_venv


def test_force(create_pex_venv):
    # type: (CreatePexVenv) -> None
    venv = create_pex_venv("--pip")
    venv.interpreter.execute(args=["-m", "pip", "install", "ansicolors==1.1.8"])
    venv.interpreter.execute(args=["-c", "import colors"])

    with pytest.raises(CalledProcessError):
        create_pex_venv()

    venv_force = create_pex_venv("--force")
    # The re-created venv should have no ansicolors installed like the prior venv.
    with pytest.raises(Executor.NonZeroExit):
        venv_force.interpreter.execute(args=["-c", "import colors"])
    # The re-created venv should have no pip installed either.
    with pytest.raises(Executor.NonZeroExit):
        venv.interpreter.execute(args=["-m", "pip", "install", "ansicolors==1.1.8"])


def execute_venv_pex_interpreter(
    venv,  # type: Virtualenv
    code=None,  # type: Optional[str]
    extra_args=(),  # type: Iterable[str]
    **extra_env  # type: Any
):
    # type: (...) -> Tuple[int, Text, Text]
    process = subprocess.Popen(
        args=[venv.join_path("pex")] + list(extra_args),
        env=make_env(PEX_INTERPRETER=True, **extra_env),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
    )
    stdout, stderr = process.communicate(input=None if code is None else code.encode())
    return process.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")


def expected_file_path(
    venv,  # type: Virtualenv
    package,  # type: str
):
    # type: (...) -> str
    return os.path.realpath(
        os.path.join(
            venv.site_packages_dir,
            os.path.sep.join(package.split(".")),
            "__init__.{ext}".format(ext="pyc" if venv.interpreter.version[0] == 2 else "py"),
        )
    )


def parse_fabric_version_output(output):
    # type: (Text) -> Dict[Text, Text]
    return dict(cast("Tuple[Text, Text]", line.split(" ", 1)) for line in output.splitlines())


def test_venv_pex(create_pex_venv):
    # type: (CreatePexVenv) -> None
    venv = create_pex_venv()
    venv_pex = venv.join_path("pex")

    fabric_output = subprocess.check_output(args=[venv_pex, "-V"])

    # N.B.: `fab -V` output looks like so:
    # $ fab -V
    # Fabric 2.5.0
    # Paramiko 2.7.2
    # Invoke 1.4.1
    versions = parse_fabric_version_output(fabric_output.decode("utf-8"))
    assert FABRIC_VERSION == versions["Fabric"]

    invoke_version = "Invoke {}".format(versions["Invoke"])
    invoke_script_output = subprocess.check_output(
        args=[venv_pex, "-V"], env=make_env(PEX_SCRIPT="invoke")
    )
    assert invoke_version == invoke_script_output.decode("utf-8").strip()

    invoke_entry_point_output = subprocess.check_output(
        args=[venv_pex, "-V"],
        env=make_env(PEX_MODULE="invoke.main:program.run"),
    )
    assert invoke_version == invoke_entry_point_output.decode("utf-8").strip()

    pex_extra_sys_path = ["/dev/null", "Bob"]
    returncode, _, stderr = execute_venv_pex_interpreter(
        venv,
        code=dedent(
            """\
            from __future__ import print_function

            import os
            import sys


            def assert_equal(test_num, expected, actual):
                if expected == actual:
                    return
                print(
                    "[{{}}] Expected {{}} but got {{}}".format(test_num, expected, actual),
                    file=sys.stderr,
                )
                sys.exit(test_num)

            assert_equal(1, {pex_extra_sys_path!r}, sys.path[-2:])

            import fabric
            assert_equal(2, {fabric!r}, os.path.realpath(fabric.__file__))

            import user.package
            assert_equal(3, {user_package!r}, os.path.realpath(user.package.__file__))
            """.format(
                pex_extra_sys_path=pex_extra_sys_path,
                fabric=expected_file_path(venv, "fabric"),
                user_package=expected_file_path(venv, "user.package"),
            )
        ),
        PEX_EXTRA_SYS_PATH=os.pathsep.join(pex_extra_sys_path),
    )
    assert 0 == returncode, stderr


def test_binary_path(create_pex_venv):
    # type: (CreatePexVenv) -> None
    code = dedent(
        """\
        import errno
        import subprocess
        import sys

        # PEXed code should be able to find all (console) scripts on the $PATH when the venv is
        # created with --bin-path set, and the scripts should all run with the venv interpreter in
        # order to find their code.

        def try_invoke(*args):
            try:
                subprocess.check_call(list(args))
                return 0
            except OSError as e:
                if e.errno == errno.ENOENT:
                    # This is what we expect when scripts are not set up on PATH via --bin-path.
                    return 1
                return 2

        exit_code = try_invoke("fab", "-V")
        exit_code += 10 * try_invoke("inv", "-V")
        exit_code += 100 * try_invoke("invoke", "-V")
        sys.exit(exit_code)
        """
    )

    venv = create_pex_venv()
    returncode, stdout, stderr = execute_venv_pex_interpreter(
        venv, code=code, PATH=tempfile.gettempdir()
    )
    assert 111 == returncode, stdout + stderr

    venv_bin_path = create_pex_venv("-f", "--bin-path", "prepend")
    returncode, _, _ = execute_venv_pex_interpreter(
        venv_bin_path, code=code, PATH=tempfile.gettempdir()
    )
    assert 0 == returncode


def test_venv_pex_interpreter_special_modes(create_pex_venv):
    # type: (CreatePexVenv) -> None
    venv = create_pex_venv()

    # special mode execute module: -m module
    returncode, stdout, stderr = execute_venv_pex_interpreter(venv, extra_args=["-m"])
    assert 2 == returncode, stderr
    assert "" == stdout

    returncode, stdout, stderr = execute_venv_pex_interpreter(
        venv, extra_args=["-m", "fabric", "--version"]
    )
    assert 0 == returncode, stderr
    versions = parse_fabric_version_output(stdout)
    assert FABRIC_VERSION == versions["Fabric"]

    # special mode execute code string: -c <str>
    returncode, stdout, stderr = execute_venv_pex_interpreter(venv, extra_args=["-c"])
    assert 2 == returncode, stderr
    assert "" == stdout

    fabric_file_code = "import fabric, os; print(os.path.realpath(fabric.__file__))"
    expected_fabric_file_path = expected_file_path(venv, "fabric")

    returncode, stdout, stderr = execute_venv_pex_interpreter(
        venv, extra_args=["-c", fabric_file_code]
    )
    assert 0 == returncode, stderr
    assert expected_fabric_file_path == stdout.strip()

    # special mode execute stdin: -
    returncode, stdout, stderr = execute_venv_pex_interpreter(
        venv, code=fabric_file_code, extra_args=["-"]
    )
    assert 0 == returncode, stderr
    assert expected_fabric_file_path == stdout.strip()

    # special mode execute python file: <py file name>
    with named_temporary_file(prefix="code", suffix=".py", mode="w") as fp:
        fp.write(fabric_file_code)
        fp.close()
        returncode, stdout, stderr = execute_venv_pex_interpreter(
            venv, code=fabric_file_code, extra_args=[fp.name]
        )
        assert 0 == returncode, stderr
        assert expected_fabric_file_path == stdout.strip()


@pytest.mark.parametrize(
    "start_method", getattr(multiprocessing, "get_all_start_methods", lambda: [None])()
)
def test_venv_multiprocessing_issues_1236(
    tmpdir,  # type: Any
    start_method,  # type: Optional[str]
):
    # type: (...) -> None
    src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src, "foo.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                def bar():
                    print('hello')
                """
            )
        )
    with safe_open(os.path.join(src, "main.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import multiprocessing
                from foo import bar

                if __name__ == '__main__':
                    if {start_method!r}:
                        multiprocessing.set_start_method({start_method!r})
                    p = multiprocessing.Process(target=bar)
                    p.start()
                """.format(
                    start_method=start_method
                )
            )
        )

    pex_file = os.path.join(str(tmpdir), "mp.pex")
    result = run_pex_command(args=["-D", src, "-m", "main", "-o", pex_file, "--include-tools"])
    result.assert_success()

    # Confirm multiprocessing works via normal PEX file execution.
    output = subprocess.check_output(args=[pex_file])
    assert "hello" == output.decode("utf-8").strip()

    # Confirm multiprocessing works via the `pex` venv script.
    venv = os.path.join(str(tmpdir), "venv")
    subprocess.check_call(args=[pex_file, "venv", venv], env=make_env(PEX_TOOLS=True))
    output = subprocess.check_output(args=[os.path.join(venv, "pex")])
    assert "hello" == output.decode("utf-8").strip()


def test_venv_symlinked_source_issues_1239(tmpdir):
    # type: (Any) -> None
    src = os.path.join(str(tmpdir), "src")
    main = os.path.join(src, "main.py")
    with safe_open(main, "w") as fp:
        fp.write("import sys; sys.exit(42)")

    pex_builder = PEXBuilder(copy_mode=CopyMode.SYMLINK)
    pex_builder.set_executable(main)
    pex_file = os.path.join(str(tmpdir), "a.pex")
    pex_builder.build(pex_file, bytecode_compile=False)
    assert 42 == subprocess.Popen(args=[pex_file]).wait()

    venv = os.path.join(str(tmpdir), "a.venv")
    subprocess.check_call(
        args=[sys.executable, "-m", "pex.tools", pex_builder.path(), "venv", venv]
    )
    venv_pex = os.path.join(venv, "pex")
    shutil.rmtree(src)
    assert 42 == subprocess.Popen(args=[venv_pex]).wait()


def test_venv_entrypoint_function_exit_code_issue_1241(tmpdir):
    # type: (Any) -> None

    pex_file = os.path.join(str(tmpdir), "ep-function.pex")
    src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src, "module.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import sys


                def target():
                    args = sys.argv[1:]
                    if args:
                        exit = args[0]
                        try:
                            return int(exit)
                        except ValueError:
                            return exit
                """
            )
        )
    result = run_pex_command(
        args=["-D", src, "-e", "module:target", "--include-tools", "-o", pex_file]
    )
    result.assert_success()

    venv = os.path.join(str(tmpdir), "ep-function.venv")
    subprocess.check_call(args=[pex_file, "venv", venv], env=make_env(PEX_TOOLS=1))

    venv_pex = os.path.join(venv, "pex")
    assert 0 == subprocess.Popen(args=[venv_pex]).wait()

    def assert_venv_process(
        args,  # type: List[str]
        expected_returncode,  # type: int
        expected_stdout="",  # type: str
        expected_stderr="",  # type: str
    ):
        # type: (...) -> None
        process = subprocess.Popen(
            args=[venv_pex] + args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        stdout, stderr = process.communicate()
        assert expected_returncode == process.returncode
        assert expected_stdout == stdout.decode("utf-8")
        assert expected_stderr == stderr.decode("utf-8")

    assert_venv_process(args=["bob"], expected_returncode=1, expected_stderr="bob\n")
    assert_venv_process(args=["42"], expected_returncode=42)


def test_venv_copies(tmpdir):
    # type: (Any) -> None

    python310 = ensure_python_interpreter(PY310)

    pex_file = os.path.join(str(tmpdir), "venv.pex")
    result = run_pex_command(args=["-o", pex_file, "--include-tools"], python=python310)
    result.assert_success()

    PEX_TOOLS = make_env(PEX_TOOLS=1)

    venv_symlinks = os.path.join(str(tmpdir), "venv.symlinks")
    subprocess.check_call(args=[python310, pex_file, "venv", venv_symlinks], env=PEX_TOOLS)
    venv_symlinks_interpreter = PythonInterpreter.from_binary(
        os.path.join(venv_symlinks, "bin", "python")
    )
    assert os.path.islink(venv_symlinks_interpreter.binary)

    venv_copies = os.path.join(str(tmpdir), "venv.copies")
    subprocess.check_call(
        args=[python310, pex_file, "venv", "--copies", venv_copies], env=PEX_TOOLS
    )
    venv_copies_interpreter = PythonInterpreter.from_binary(
        os.path.join(venv_copies, "bin", "python")
    )
    assert not os.path.islink(venv_copies_interpreter.binary)


def test_relocatable_venv(tmpdir):
    # type: (Any) -> None

    pex_file = os.path.join(str(tmpdir), "relocatable.pex")
    src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src, "main.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import sys
                from colors import blue


                print(blue(sys.executable))
                """
            )
        )
    result = run_pex_command(
        args=["-D", src, "ansicolors==1.1.8", "-m", "main", "--include-tools", "-o", pex_file]
    )
    result.assert_success()

    venv = os.path.join(str(tmpdir), "relocatable.venv")
    subprocess.check_call(args=[pex_file, "venv", venv], env=make_env(PEX_TOOLS=1))
    subprocess.check_call(args=[os.path.join(venv, "pex")])

    relocated_relpath = "relocated.venv"
    relocated_venv = os.path.join(str(tmpdir), relocated_relpath)

    # Since the venv pex script contains a shebang with an absolute path to the venv python
    # interpreter, a move of the venv makes the script un-runnable directly.
    shutil.move(venv, relocated_venv)
    with pytest.raises(OSError) as exec_info:
        subprocess.check_call(args=[os.path.join(relocated_venv, "pex")])
    assert errno.ENOENT == exec_info.value.errno

    # But we should be able to run the script using the moved venv's interpreter.
    subprocess.check_call(
        args=[
            os.path.join(relocated_relpath, "bin", "python"),
            os.path.join(relocated_relpath, "pex"),
        ],
        cwd=str(tmpdir),
    )


def test_compile(tmpdir):
    # type: (Any) -> None

    def collect_files(
        root_dir,  # type: str
        extension,  # type: str
    ):
        # type: (...) -> Set[str]
        return {
            os.path.relpath(os.path.join(root, f), root_dir)
            for root, _, files in os.walk(root_dir, followlinks=False)
            for f in files
            if f.endswith(extension)
        }

    pex_file = os.path.join(str(tmpdir), "compile.pex")
    src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src, "main.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                from colors import yellow

            
                print(yellow("Slartibartfast"))
                """
            )
        )
    result = run_pex_command(
        args=["-D", src, "ansicolors==1.0.2", "-m", "main", "--include-tools", "-o", pex_file]
    )
    result.assert_success()

    venv = os.path.join(str(tmpdir), "venv")
    subprocess.check_call(args=[pex_file, "venv", venv], env=make_env(PEX_TOOLS=1))
    # N.B.: The right way to discover the site-packages dir is via site.getsitepackages().
    # Unfortunately we use an old version of virtualenv to create PyPy <= 3.7 and CPython 2.7 venvs
    # and it does not add a getsitepackages function to site.py; so we cheat.
    if IS_PYPY and PY_VER <= (3, 7):
        site_packages = "site-packages"
    else:
        site_packages = os.path.join(
            "lib",
            "{python}{major}.{minor}".format(
                python="pypy" if IS_PYPY else "python",
                major=sys.version_info[0],
                minor=sys.version_info[1],
            ),
            "site-packages",
        )

    # Ensure we have at least the basic direct dependency python files we expect.
    venv_py_files = collect_files(venv, ".py")
    assert os.path.join(site_packages, "main.py") in venv_py_files
    assert os.path.join(site_packages, "colors.py") in venv_py_files
    assert "__main__.py" in venv_py_files

    compile_venv = os.path.join(str(tmpdir), "compile.venv")
    subprocess.check_call(
        args=[pex_file, "venv", "--compile", compile_venv], env=make_env(PEX_TOOLS=1)
    )
    # Ensure all original py files have a compiled counterpart.
    for py_file in venv_py_files:
        if PY2:
            assert os.path.exists(os.path.join(compile_venv, py_file + "c"))
        else:
            name, _ = os.path.splitext(os.path.basename(py_file))
            assert os.path.exists(
                os.path.join(
                    compile_venv,
                    os.path.dirname(py_file),
                    "__pycache__",
                    "{name}.{cache_tag}.pyc".format(
                        name=name, cache_tag=sys.implementation.cache_tag
                    ),
                )
            )

    compile_venv_pyc_files = collect_files(compile_venv, ".pyc")
    subprocess.check_call(args=[os.path.join(compile_venv, "pex")])
    assert compile_venv_pyc_files == collect_files(
        compile_venv, ".pyc"
    ), "Expected no new compiled python files."


def test_strip_pex_env(tmpdir):
    # type: (Any) -> None

    def create_pex_venv(strip_pex_env):
        # type: (bool) -> str
        pex = os.path.join(str(tmpdir), "strip_{}.pex".format(strip_pex_env))
        run_pex_command(
            args=[
                "--strip-pex-env" if strip_pex_env else "--no-strip-pex-env",
                "--include-tools",
                "-o",
                pex,
            ]
        ).assert_success()

        venv = os.path.join(str(tmpdir), "strip_{}.venv".format(strip_pex_env))
        subprocess.check_call(args=[pex, "venv", venv], env=make_env(PEX_TOOLS=1))
        return venv

    check_pex_env_vars_code = dedent(
        """\
        from __future__ import print_function

        import os
        import sys


        pex_env_vars = 0
        for name, value in os.environ.items():
            if name.startswith("PEX_"):
                pex_env_vars += 1
                print(
                    "Un-stripped: {name}={value}".format(name=name, value=value), file=sys.stderr
                )
        sys.exit(pex_env_vars)
        """
    )

    two_pex_env_vars = {
        name: value
        for name, value in make_env(PEX_ROOT="42", PEX_TOOLS=1).items()
        if name in ("PEX_ROOT", "PEX_TOOLS") or not name.startswith("PEX_")
    }
    assert 2 == len([name for name in two_pex_env_vars if name.startswith("PEX_")])

    strip_venv = create_pex_venv(strip_pex_env=True)
    subprocess.check_call(
        args=[os.path.join(strip_venv, "pex"), "-c", check_pex_env_vars_code], env=two_pex_env_vars
    )

    no_strip_venv = create_pex_venv(strip_pex_env=False)
    process = subprocess.Popen(
        args=[os.path.join(no_strip_venv, "pex"), "-c", check_pex_env_vars_code],
        env=two_pex_env_vars,
    )
    assert 2 == process.wait()


def test_warn_unused_pex_env_vars():
    # type: () -> None
    # N.B.: We don't use the pytest tmpdir fixture here since it creates fairly length paths under
    # /tmp and under macOS, where TMPDIR is already fairly deeply nested, we trigger Pex warnings
    # about script shebang length. Those warnings pollute stderr.
    tmpdir = safe_mkdtemp()
    venv_pex = os.path.join(tmpdir, "venv.pex")
    run_pex_command(["--venv", "-o", venv_pex]).assert_success()

    def assert_execute_venv_pex(expected_stderr, **env_vars):
        env = os.environ.copy()
        env.update(env_vars)
        process = subprocess.Popen(
            [venv_pex, "-c", ""], stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env
        )
        stdout, stderr = process.communicate()
        assert 0 == process.returncode
        assert not stdout
        assert expected_stderr.strip() == stderr.decode("utf-8").strip()

    assert_execute_venv_pex(expected_stderr="")
    assert_execute_venv_pex(expected_stderr="", PEX_ROOT=os.path.join(tmpdir, "pex_root"))
    assert_execute_venv_pex(expected_stderr="", PEX_VENV="1")
    assert_execute_venv_pex(expected_stderr="", PEX_EXTRA_SYS_PATH="more")
    assert_execute_venv_pex(expected_stderr="", PEX_VERBOSE="0")

    assert_execute_venv_pex(
        expected_stderr=dedent(
            """\
            Ignoring the following environment variables in Pex venv mode:
            PEX_INHERIT_PATH=fallback
            """
        ),
        PEX_INHERIT_PATH="fallback",
    )

    assert_execute_venv_pex(
        expected_stderr=dedent(
            """\
            Ignoring the following environment variables in Pex venv mode:
            PEX_COVERAGE=1
            PEX_INHERIT_PATH=fallback
            """
        ),
        PEX_COVERAGE="1",
        PEX_INHERIT_PATH="fallback",
        PEX_VERBOSE="0",
    )


def test_custom_prompt(tmpdir):
    # type: (Any) -> None
    pex_root = os.path.join(str(tmpdir), "pex_root")
    venv_pex = os.path.join(str(tmpdir), "venv.pex")
    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "-o",
            venv_pex,
            "--include-tools",
        ]
    ).assert_success()

    venv_dir = os.path.join(str(tmpdir), "venv_dir")
    subprocess.check_call(
        args=[venv_pex, "venv", "--prompt", "jane", venv_dir], env=make_env(PEX_TOOLS=True)
    )

    if PY_VER == (2, 7) or (IS_PYPY and PY_VER <= (3, 7)):
        # Neither CPython 2.7 not PyPy interpreters have (functioning) venv modules; so we create
        # their venvs with an old copy of virtualenv that does not surround the prompt with parens.
        expected_prompt = "jane"
    elif PY_VER == (3, 5):
        # We can't set the prompt for CPython 3.5 so we expect the name of the venv dir.
        expected_prompt = "(venv_dir)"
    else:
        expected_prompt = "(jane)"

    output = subprocess.check_output(
        args=[
            "/usr/bin/env",
            "bash",
            "-c",
            "source {} && echo $PS1".format(os.path.join(venv_dir, "bin", "activate")),
        ],
        env=make_env(TERM="dumb", COLS=80),
    )
    assert expected_prompt == output.decode("utf-8").strip()


@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
def test_remove(
    tmpdir,
    layout,  # type: Layout.Value
):
    # type: (...) -> None
    pex_root = os.path.join(str(tmpdir), "pex_root")

    def create_venv_pex():
        # type: () -> str
        venv_pex = os.path.join(str(tmpdir), "venv.pex")
        run_pex_command(
            args=[
                "--pex-root",
                pex_root,
                "--runtime-pex-root",
                pex_root,
                "-o",
                venv_pex,
                "--include-tools",
            ]
        ).assert_success()
        return venv_pex

    venv_dir = os.path.join(str(tmpdir), "venv_dir")
    assert not os.path.exists(venv_dir)

    venv_pex = create_venv_pex()
    subprocess.check_call(args=[venv_pex, "venv", venv_dir], env=make_env(PEX_TOOLS=True))
    assert os.path.exists(venv_dir)
    assert os.path.exists(venv_pex)
    assert os.path.exists(pex_root)

    shutil.rmtree(venv_dir)
    assert not os.path.exists(venv_dir)

    subprocess.check_call(
        args=[venv_pex, "venv", "--rm", "pex", venv_dir], env=make_env(PEX_TOOLS=True)
    )
    assert os.path.exists(venv_dir)
    assert not os.path.exists(venv_pex)
    assert os.path.exists(pex_root)

    shutil.rmtree(venv_dir)
    assert not os.path.exists(venv_dir)
    venv_pex = create_venv_pex()

    subprocess.check_call(
        args=[venv_pex, "venv", "--rm", "all", venv_dir], env=make_env(PEX_TOOLS=True)
    )
    assert os.path.exists(venv_dir)
    assert not os.path.exists(venv_pex)
    assert not os.path.exists(pex_root)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import re

import pytest

from pex.resolve.path_mappings import PathMapping, PathMappings
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Tuple


def create_path_mappings(*mappings):
    # type: (*Tuple[str, str]) -> PathMappings
    return PathMappings(tuple(PathMapping(path=path, name=name) for path, name in mappings))


def test_invalid():
    # type: () -> None

    with pytest.raises(ValueError, match=re.escape("Mapped paths must be absolute. Given: foo")):
        create_path_mappings(("./foo", "A"))


def test_normalize():
    # type: () -> None

    def check_path(
        path_mappings,  # type: PathMappings
        expected_path,  # type: str
    ):
        # type: (...) -> PathMappings
        assert 1 == len(path_mappings.mappings)
        assert expected_path == path_mappings.mappings[0].path
        return path_mappings

    assert check_path(create_path_mappings(("/tmp/foo", "A")), "/tmp/foo") == check_path(
        create_path_mappings(("/tmp/./foo/", "A")), "/tmp/foo"
    )


def test_noop():
    # type: () -> None

    path_mappings = create_path_mappings(("/tmp/foo", "A"))

    assert "foo" == path_mappings.maybe_canonicalize("foo")
    assert "/tmp/bar" == path_mappings.maybe_canonicalize("/tmp/bar")

    assert "foo" == path_mappings.maybe_reify("foo")
    assert "/tmp/foo" == path_mappings.maybe_reify("/tmp/foo")
    assert "A" == path_mappings.maybe_reify("A")
    assert "$A" == path_mappings.maybe_reify("$A")


def test_canonicalize():
    # type: () -> None

    path_mappings = create_path_mappings(("/tmp/foo", "A"), ("/tmp/bar/", "B"))

    assert "${A}" == path_mappings.maybe_canonicalize("/tmp/foo")
    assert "${B}/" == path_mappings.maybe_canonicalize("/tmp/bar/")
    assert "${A}/bar" == path_mappings.maybe_canonicalize("/tmp/foo/bar")

    assert "file://${A}/bar" == path_mappings.maybe_canonicalize("file:///tmp/foo/bar")
    assert "baz @ file://${B}/baz" == path_mappings.maybe_canonicalize("baz @ file:///tmp/bar/baz")


def test_reify():
    # type: () -> None

    path_mappings = create_path_mappings(("/tmp/foo", "A"), ("/tmp/bar/", "B"))

    assert "/tmp/foo" == path_mappings.maybe_reify("${A}")
    assert "/tmp/bar/" == path_mappings.maybe_reify("${B}/")
    assert "/tmp/foo/bar" == path_mappings.maybe_reify("${A}/bar")

    assert "file:///tmp/foo/bar" == path_mappings.maybe_reify("file://${A}/bar")
    assert "baz @ file:///tmp/bar/baz" == path_mappings.maybe_reify("baz @ file://${B}/baz")



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pex.sorted_tuple import SortedTuple


def test_empty():
    # type: () -> None
    empty = SortedTuple()
    assert 0 == len(empty)
    assert [] == list(empty)


def test_non_empty():
    # type: () -> None
    non_empty = SortedTuple([1])
    assert [1] == list(non_empty)


def test_sorting():
    # type: () -> None
    sorted_tuple = SortedTuple([3, 1, 5, 1, 3, 2, 4, 0])
    assert [0, 1, 1, 2, 3, 3, 4, 5] == list(sorted_tuple)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess

from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_empty_pex_path(tmpdir):
    # type: (Any) -> None

    empty_pex = os.path.join(str(tmpdir), "empty.pex")
    run_pex_command(args=["-o", empty_pex]).assert_success()

    # The previously failing case.
    subprocess.check_call(args=[empty_pex, "-c", ""], env=make_env(PEX_PATH=""))

    colors_pex = os.path.join(str(tmpdir), "colors.pex")
    run_pex_command(
        args=["ansicolors==1.1.8", "-o", colors_pex, "--layout", "packed"]
    ).assert_success()

    assert 0 != subprocess.call(
        args=[empty_pex, "-c", "import colors"]
    ), "Expected a PEX_PATH including colors.pex to be needed in order to import colors."
    subprocess.check_call(
        args=[empty_pex, "-c", "import colors"], env=make_env(PEX_PATH=colors_pex)
    )
    # Demonstrate that . can be used instead of the empty string to trigger a CWD PEX_PATH if that
    # is what is intended.
    subprocess.check_call(
        args=[empty_pex, "-c", "import colors"], env=make_env(PEX_PATH="."), cwd=colors_pex
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
import shutil

from pex import hashing
from pex.atomic_directory import atomic_directory
from pex.common import safe_mkdir, safe_mkdtemp
from pex.compatibility import url_unquote, urlparse
from pex.hashing import Sha256
from pex.jobs import Job, Raise, SpawnedJob, execute_parallel
from pex.pip.download_observer import DownloadObserver
from pex.pip.installation import get_pip
from pex.pip.tool import PackageIndexConfiguration, Pip
from pex.resolve import locker
from pex.resolve.locked_resolve import Artifact, FileArtifact, LockConfiguration, LockStyle
from pex.resolve.resolved_requirement import Fingerprint, PartialArtifact
from pex.resolve.resolvers import Resolver
from pex.result import Error
from pex.targets import LocalInterpreter, Target
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import Dict, Iterable, Iterator, Optional, Union

    import attr  # vendor:skip

    from pex.hashing import HintedDigest
else:
    from pex.third_party import attr


_DOWNLOADS_DIRS = {}  # type: Dict[str, str]


def get_downloads_dir(pex_root=None):
    # type: (Optional[str]) -> str
    root_dir = pex_root or ENV.PEX_ROOT
    downloads_dir = _DOWNLOADS_DIRS.get(root_dir)
    if downloads_dir is None:
        downloads_dir = os.path.join(root_dir, "downloads")
        safe_mkdir(downloads_dir)
        _DOWNLOADS_DIRS[root_dir] = downloads_dir
    return downloads_dir


@attr.s(frozen=True)
class ArtifactDownloader(object):
    resolver = attr.ib()  # type: Resolver
    target = attr.ib(default=LocalInterpreter.create())  # type: Target
    package_index_configuration = attr.ib(
        default=PackageIndexConfiguration.create()
    )  # type: PackageIndexConfiguration
    max_parallel_jobs = attr.ib(default=None)  # type: Optional[int]
    pip = attr.ib(init=False)  # type: Pip

    @pip.default
    def _pip(self):
        return get_pip(
            interpreter=self.target.get_interpreter(),
            version=self.package_index_configuration.pip_version,
            resolver=self.resolver,
        )

    @staticmethod
    def _fingerprint_and_move(path):
        # type: (str) -> Fingerprint
        digest = Sha256()
        hashing.file_hash(path, digest)
        fingerprint = digest.hexdigest()
        target_dir = os.path.join(get_downloads_dir(), fingerprint)
        with atomic_directory(target_dir) as atomic_dir:
            if not atomic_dir.is_finalized():
                shutil.move(path, os.path.join(atomic_dir.work_dir, os.path.basename(path)))
        return Fingerprint.from_hashing_fingerprint(fingerprint)

    @staticmethod
    def _create_file_artifact(
        url,  # type: str
        fingerprint,  # type: Fingerprint
        verified,  # type: bool
    ):
        # type: (...) -> FileArtifact
        fingerprinted_artifact = Artifact.from_url(url, fingerprint, verified=verified)
        if not isinstance(fingerprinted_artifact, FileArtifact):
            raise ValueError(
                "Expected a file artifact, given url {url} which is a {artifact}.".format(
                    url=url, artifact=fingerprinted_artifact
                )
            )
        return fingerprinted_artifact

    def _download(
        self,
        url,  # type: str
        download_dir,  # type: str
    ):
        # type: (...) -> Job

        for password_entry in self.package_index_configuration.password_entries:
            credentialed_url = password_entry.maybe_inject_in_url(url)
            if credentialed_url:
                url = credentialed_url
                break

        # Although we don't actually need to observe the download, we do need to patch Pip to not
        # care about wheel tags, environment markers or Requires-Python. The locker's download
        # observer does just this for universal locks with no target system or requires python
        # restrictions.
        download_observer = DownloadObserver(
            analyzer=None,
            patch_set=locker.patch(lock_configuration=LockConfiguration(style=LockStyle.UNIVERSAL)),
        )
        return self.pip.spawn_download_distributions(
            download_dir=download_dir,
            requirements=[url],
            transitive=False,
            package_index_configuration=self.package_index_configuration,
            observer=download_observer,
        )

    def _download_and_fingerprint(self, url):
        # type: (str) -> SpawnedJob[FileArtifact]
        downloads = get_downloads_dir()
        download_dir = safe_mkdtemp(prefix="fingerprint_artifact.", dir=downloads)

        url_info = urlparse.urlparse(url)
        src_file = url_unquote(url_info.path)
        temp_dest = os.path.join(download_dir, os.path.basename(src_file))

        if url_info.scheme == "file":
            shutil.copy(src_file, temp_dest)
            return SpawnedJob.completed(
                self._create_file_artifact(
                    url, fingerprint=self._fingerprint_and_move(temp_dest), verified=True
                )
            )

        return SpawnedJob.and_then(
            self._download(url=url, download_dir=download_dir),
            result_func=lambda: self._create_file_artifact(
                url, fingerprint=self._fingerprint_and_move(temp_dest), verified=True
            ),
        )

    def _to_file_artifact(self, artifact):
        # type: (PartialArtifact) -> SpawnedJob[FileArtifact]
        url = artifact.url.normalized_url
        fingerprint = artifact.fingerprint
        if fingerprint:
            return SpawnedJob.completed(
                self._create_file_artifact(url, fingerprint, verified=artifact.verified)
            )
        return self._download_and_fingerprint(url)

    def fingerprint(self, artifacts):
        # type: (Iterable[PartialArtifact]) -> Iterator[FileArtifact]
        return execute_parallel(
            inputs=artifacts,
            spawn_func=self._to_file_artifact,
            error_handler=Raise[PartialArtifact, FileArtifact](IOError),
        )

    def download(
        self,
        artifact,  # type: FileArtifact
        dest_dir,  # type: str
        digest,  # type: HintedDigest
    ):
        # type: (...) -> Union[str, Error]
        dest_file = os.path.join(dest_dir, artifact.filename)

        url_info = urlparse.urlparse(artifact.url)
        if url_info.scheme == "file":
            src_file = url_unquote(url_info.path)
            try:
                shutil.copy(src_file, dest_file)
            except (IOError, OSError) as e:
                return Error(str(e))
        else:
            try:
                self._download(url=artifact.url, download_dir=dest_dir).wait()
            except Job.Error as e:
                return Error((e.stderr or str(e)).splitlines()[-1])
        hashing.file_hash(dest_file, digest)
        return artifact.filename



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os.path
from textwrap import dedent

from pex.cli.testing import run_pex3
from pex.common import safe_open, touch
from pex.compatibility import commonpath
from pex.interpreter import PythonInterpreter
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_missing_download_lock_analysis_handling(
    tmpdir,  # type: Any
    py310,  # type: PythonInterpreter
):
    # type: (...) -> None

    my_feast = os.path.join(str(tmpdir), "intermediary")
    touch(os.path.join(my_feast, "README.rst"))
    with safe_open(os.path.join(my_feast, "pyproject.toml"), "w") as fp:
        fp.write(
            dedent(
                """\
                [build-system]
                requires = ["setuptools", "wheel"]
                build-backend = "setuptools.build_meta"

                [project]
                name = "my_feast"
                version = "0.0.1"
                authors = [
                    {name = "John Sirois", email = "john.sirois@gmail.com"},
                ]
                description = "Simulates the more complex and expensive 'feast' in the issue OP."
                readme = "README.rst"
                requires-python = ">=3.7"
                license = {text = "BSD-3-Clause"}
                dependencies = [
                    "SQLAlchemy[mypy]>1,<2",
                ]
                """
            )
        )

    pex_root = os.path.join(str(tmpdir), "pex_root")
    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock",
        "create",
        "--pex-root",
        pex_root,
        "--python-path",
        py310.binary,
        "--interpreter-constraint",
        "==3.10.*",
        "--style",
        "universal",
        "--resolver-version",
        "pip-2020-resolver",
        "--target-system",
        "linux",
        "--target-system",
        "mac",
        my_feast,
        "sqlalchemy==1.3.24",
        "--indent",
        "2",
        "-o",
        lock,
    ).assert_success()

    result = run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--lock",
            lock,
            "sqlalchemy",
            "--",
            "-c",
            dedent(
                """\
                import json
                import sys

                import sqlalchemy


                json.dump(
                    {"version": sqlalchemy.__version__, "file": sqlalchemy.__file__},
                    sys.stdout,
                )
                """
            ),
        ],
        python=py310.binary,
    )
    result.assert_success()

    data = json.loads(result.output)
    assert "1.3.24" == data["version"]
    assert pex_root == commonpath([pex_root, data["file"]])



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import errno
import logging
import os
from argparse import ArgumentParser

from pex import pex_warnings
from pex.common import safe_delete, safe_rmtree
from pex.enum import Enum
from pex.executor import Executor
from pex.pex import PEX
from pex.result import Ok, Result, try_
from pex.tools.command import PEXCommand
from pex.typing import TYPE_CHECKING
from pex.venv import installer, installer_options
from pex.venv.install_scope import InstallScope
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Optional

    import attr  # vendor:skip
else:
    from pex.third_party import attr


logger = logging.getLogger(__name__)


class RemoveScope(Enum["RemoveScope.Value"]):
    class Value(Enum.Value):
        pass

    PEX = Value("pex")
    PEX_AND_PEX_ROOT = Value("all")


@attr.s(frozen=True)
class InstallScopeState(object):
    @classmethod
    def load(cls, venv_dir):
        # type: (str) -> InstallScopeState

        state_file = os.path.join(venv_dir, ".pex-venv-scope")
        prior_state = None  # type: Optional[InstallScope.Value]
        try:
            with open(state_file) as fp:
                prior_state = InstallScope.for_value(fp.read().strip())
        except IOError as e:
            if e.errno != errno.ENOENT:
                raise e

        return cls(venv_dir=venv_dir, state_file=state_file, prior_state=prior_state)

    venv_dir = attr.ib()  # type: str
    _state_file = attr.ib()  # type: str
    _prior_state = attr.ib(default=None)  # type: Optional[InstallScope.Value]

    @property
    def is_partial_install(self):
        return self._prior_state in (InstallScope.DEPS_ONLY, InstallScope.SOURCE_ONLY)

    def save(self, install_scope):
        # type: (InstallScope.Value) -> None
        if {InstallScope.DEPS_ONLY, InstallScope.SOURCE_ONLY} == {self._prior_state, install_scope}:
            install_scope = InstallScope.ALL
        with open(self._state_file, "w") as fp:
            fp.write(str(install_scope))


class Venv(PEXCommand):
    """Creates a venv from the PEX file."""

    @classmethod
    def add_arguments(cls, parser):
        # type: (ArgumentParser) -> None
        parser.add_argument(
            "venv",
            nargs=1,
            metavar="PATH",
            help="The directory to create the virtual environment in.",
        )
        installer_options.register(parser, include_force_switch=True)
        parser.add_argument(
            "--rm",
            "--remove",
            dest="remove",
            default=None,
            choices=RemoveScope.values(),
            type=RemoveScope.for_value,
            help=(
                "Remove the PEX after creating a venv from it if the {pex!r} value is specified; "
                "otherwise, remove the PEX and the PEX_ROOT if the {all!r} value is "
                "specified.".format(
                    pex=RemoveScope.PEX.value, all=RemoveScope.PEX_AND_PEX_ROOT.value
                )
            ),
        )
        cls.register_global_arguments(parser, include_verbosity=False)

    def run(self, pex):
        # type: (PEX) -> Result

        installer_configuration = installer_options.configure(self.options)

        venv_dir = self.options.venv[0]
        install_scope_state = InstallScopeState.load(venv_dir)
        if install_scope_state.is_partial_install and not installer_configuration.force:
            venv = Virtualenv(venv_dir)
        else:
            venv = Virtualenv.create(
                venv_dir,
                interpreter=pex.interpreter,
                force=installer_configuration.force,
                copies=installer_configuration.copies,
                prompt=installer_configuration.prompt,
            )

        if installer_configuration.pip:
            try_(
                installer.ensure_pip_installed(
                    venv,
                    distributions=tuple(pex.resolve()),
                    scope=installer_configuration.scope,
                    collisions_ok=installer_configuration.collisions_ok,
                    source="PEX at {pex}".format(pex=pex.path()),
                )
            )

        if installer_configuration.prompt != venv.custom_prompt:
            logger.warning(
                "Unable to apply custom --prompt {prompt!r} in {python} venv; continuing with the "
                "default prompt.".format(
                    prompt=installer_configuration.prompt, python=venv.interpreter.identity
                )
            )
        installer.populate_venv_from_pex(
            venv,
            pex,
            bin_path=installer_configuration.bin_path,
            collisions_ok=installer_configuration.collisions_ok,
            symlink=False,
            scope=installer_configuration.scope,
            hermetic_scripts=installer_configuration.hermetic_scripts,
        )

        if installer_configuration.compile:
            try:
                pex.interpreter.execute(["-m", "compileall", venv_dir])
            except Executor.NonZeroExit as non_zero_exit:
                pex_warnings.warn("ignoring compile error {}".format(repr(non_zero_exit)))

        if self.options.remove is not None:
            if os.path.isdir(pex.path()):
                safe_rmtree(pex.path())
            else:
                safe_delete(pex.path())
            if self.options.remove is RemoveScope.PEX_AND_PEX_ROOT:
                safe_rmtree(pex.pex_info().pex_root)

        install_scope_state.save(installer_configuration.scope)
        return Ok()



# Copyright 2018 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import os
import pkgutil
import shutil
import subprocess
import sys
from argparse import ArgumentParser
from collections import OrderedDict, defaultdict

from colors import bold, green, yellow
from redbaron import CommentNode, LiteralyEvaluable, NameNode, RedBaron

from pex.common import safe_delete, safe_mkdir, safe_mkdtemp, safe_open, safe_rmtree
from pex.vendor import VendorSpec, iter_vendor_specs


class ImportRewriter(object):
    """Rewrite imports of a set of root modules to be prefixed.

    Rewriting imports in this way is often referred to as shading. In combination with a PEP-302
    importer that can keep shaded code isolated from the normal ``sys.path`` robust vendoring of
    third party code can be achieved.
    """

    @staticmethod
    def _parse(python_file):
        with open(python_file) as fp:
            # NB: RedBaron is used instead of ``ast`` since it can round-trip from source code without
            # losing formatting. See: https://github.com/PyCQA/redbaron
            return RedBaron(fp.read())

    @staticmethod
    def _skip(node):
        next_node = node.next_recursive
        if isinstance(next_node, CommentNode) and next_node.value.strip() == "# vendor:skip":
            print("Skipping {} as directed by {}".format(node, next_node))
            return True
        return False

    @staticmethod
    def _find_literal_node(statement, call_argument):
        # The list of identifiers is large and they represent disjoint types:
        #   'StringNode'
        #   'BinaryStringNode'
        #   'RawStringNode'
        #   'InterpolatedRawStringNode'
        #   ...
        # Instead of trying to keep track of that, since we're specialized to the context of __import__,
        # Just accept any LiteralyEvaluable node (a mixin the above all implement) that python
        # accepts - except NameNode which we don't want and whose existence as a LiteralyEvaluable is
        # questionable to boot. In other words, we do not want to attempt to transform:
        #   variable = 'bob'
        #   __import__(variable, ...)
        # We just accept we'll miss more complex imports like this and have to fix by hand.

        if isinstance(call_argument.value, NameNode):
            print(yellow("WARNING: Skipping {}".format(statement)), file=sys.stderr)
        elif isinstance(call_argument.value, LiteralyEvaluable):
            return call_argument.value

    @staticmethod
    def _modify_import(original, modified):
        indent = " " * (modified.absolute_bounding_box.top_left.column - 1)
        return os.linesep.join(
            indent + line
            for line in (
                'if "__PEX_UNVENDORED__" in __import__("os").environ:',
                "  {}  # vendor:skip".format(original),
                "else:",
                "  {}".format(modified),
            )
        )

    @classmethod
    def for_path_items(cls, prefix, path_items):
        pkg_names = frozenset(pkg_name for _, pkg_name, _ in pkgutil.iter_modules(path=path_items))
        return cls(prefix=prefix, packages=pkg_names)

    def __init__(self, prefix, packages):
        self._prefix = prefix
        self._packages = packages

    def rewrite(self, python_file):
        modififications = OrderedDict()

        red_baron = self._parse(python_file)
        modififications.update(self._modify__import__calls(red_baron))
        modififications.update(self._modify_import_statements(red_baron))
        modififications.update(self._modify_from_import_statements(red_baron))

        if modififications:
            with open(python_file, "w") as fp:
                fp.write(red_baron.dumps())
            return modififications

    def _modify__import__calls(self, red_baron):  # noqa: We want __import__ as part of the name.
        for call_node in red_baron.find_all("CallNode"):
            if call_node.previous and call_node.previous.value == "__import__":
                if self._skip(call_node):
                    continue

                parent = call_node.parent_find("AtomtrailersNode")
                original = parent.copy()
                first_argument = call_node[0]
                raw_value = self._find_literal_node(parent, first_argument)
                if raw_value:
                    value = raw_value.to_python()
                    root_package = value.split(".")[0]
                    if root_package in self._packages:
                        raw_value.replace("{!r}".format(self._prefix + "." + value))

                        parent.replace(self._modify_import(original, parent))
                        yield original, parent

    def _modify_import_statements(self, red_baron):
        for import_node in red_baron.find_all("ImportNode"):
            modified = False
            if self._skip(import_node):
                continue

            original = import_node.copy()
            for index, import_module in enumerate(import_node):
                root_package = import_module[0]
                if root_package.value not in self._packages:
                    continue

                # We need to handle 4 possible cases:
                # 1. a -> pex.third_party.a as a
                # 2. a.b -> pex.third_party.a.b, pex.third_party.a as a
                # 3. a as b -> pex.third_party.a as b
                # 4. a.b as c -> pex.third_party.a.b as c
                #
                # Of these, 2 is the interesting case. The code in question would be like:
                # ```
                # import a.b.c
                # ...
                # a.b.c.func()
                # ```
                # So we need to have imported `a.b.c` but also exposed the root of that package path, `a`
                # under the name expected by code. The import of the `a.b.c` leaf ensures all parent
                # packages have been imported (getting the middle `b` in this case which is not explicitly
                # imported). This ensures the code can traverse from the re-named root - `a` in this
                # example, through middle nodes (`a.b`) all the way to the leaf target (`a.b.c`).

                modified = True

                def prefixed_fullname():
                    return "{prefix}.{module}".format(
                        prefix=self._prefix, module=".".join(map(str, import_module))
                    )

                if import_module.target:  # Cases 3 and 4.
                    import_module.value = prefixed_fullname()
                else:
                    if len(import_module) > 1:  # Case 2.
                        import_node.insert(index, prefixed_fullname())

                    # Cases 1 and 2.
                    import_module.value = "{prefix}.{root}".format(
                        prefix=self._prefix, root=root_package.value
                    )
                    import_module.target = root_package.value

            if modified:
                import_node.replace(self._modify_import(original, import_node))
                yield original, import_node

    def _modify_from_import_statements(self, red_baron):
        for from_import_node in red_baron.find_all("FromImportNode"):
            if self._skip(from_import_node):
                continue

            if len(from_import_node) == 0:
                # NB: `from . import ...` has length 0, but we don't care about relative imports which will
                # point back into vendored code if the origin is within vendored code.
                continue

            original = from_import_node.copy()
            root_package = from_import_node[0]
            if root_package.value in self._packages:
                root_package.replace(
                    "{prefix}.{root}".format(prefix=self._prefix, root=root_package.value)
                )

                from_import_node.replace(self._modify_import(original, from_import_node))
                yield original, from_import_node


class VendorizeError(Exception):
    """Indicates an error was encountered updating vendored libraries."""


def find_site_packages(prefix_dir):
    for root, dirs, _ in os.walk(prefix_dir):
        for d in dirs:
            if "site-packages" == d:
                return os.path.join(root, d)

    raise VendorizeError(
        "Failed to locate a site-packages directory within installation prefix "
        "{prefix_dir}.".format(prefix_dir=prefix_dir)
    )


def vendorize(root_dir, vendor_specs, prefix, update):
    # There is bootstrapping catch-22 here. In order for `pex.third_party` to work, all 3rdparty
    # importable code must lie at the top of its vendored chroot. Although
    # `pex.pep_376.Record.fixup_install` encodes the logic to achieve this layout, we can't run
    # that without 1st approximating that layout!. We take the tack of doing the --prefix
    # install off into a temp dir, moving the site-packages importables into the vendor chroot,
    # importing the code we'll need, then moving the importables back.
    moved = {}

    prefix_dir_by_vendor_spec = defaultdict(safe_mkdtemp)
    for vendor_spec in vendor_specs:
        # NB: We set --no-build-isolation to prevent pip from installing the requirements listed in
        # its [build-system] config in its pyproject.toml.
        #
        # Those requirements are (currently) the unpinned ["setuptools", "wheel"], which will cause
        # pip to use the latest version of those packages.  This is a hermeticity problem:
        # re-vendoring a package at a later time may yield a different result.  At the very least
        # the result will differ in the version embedded in the WHEEL metadata file, which causes
        # issues with our tests.
        #
        # Setting --no-build-isolation means that versions of setuptools and wheel must be provided
        # in the environment in which we run the pip command, which is the environment in which we
        # run pex.vendor. Since we document that pex.vendor should be run via tox, that environment
        # will contain pinned versions of setuptools and wheel. We further set `--no-cache-dir` so
        # that Pip finds no newer versions of wheel in its cache. As a result, vendoring (at least
        # via tox) is hermetic.
        requirement = vendor_spec.prepare()
        cmd = [
            "pip",
            "install",
            "--no-build-isolation",
            "--no-cache-dir",
            "--no-compile",
            "--prefix",
            prefix_dir_by_vendor_spec[vendor_spec],
            # In `--prefix` scheme, Pip warns about installed scripts not being on $PATH. We fix
            # this when a PEX is turned into a venv.
            "--no-warn-script-location",
            # In `--prefix` scheme, Pip normally refuses to install a dependency already in the
            # `sys.path` of Pip itself since the requirement is already satisfied. Since `pip`,
            # `setuptools` and `wheel` are always in that `sys.path` (Our `pip.pex` venv PEX), we
            # force installation so that PEXes with dependencies on those projects get them properly
            # installed instead of skipped.
            "--force-reinstall",
            "--ignore-installed",
            requirement,
        ]

        constraints_file = os.path.join(vendor_spec.target_dir, "constraints.txt")
        if update and (vendor_spec.constrain or vendor_spec.constraints):
            safe_delete(constraints_file)
            if vendor_spec.constraints:
                with safe_open(constraints_file, "w") as fp:
                    for constraint in vendor_spec.constraints:
                        print(constraint, file=fp)
                cmd.extend(["--constraint", constraints_file])
        elif vendor_spec.constrain:
            # Use the last checked-in constraints if any.
            subprocess.call(["git", "checkout", "--", constraints_file])
            if os.path.isfile(constraints_file):
                cmd.extend(["--constraint", constraints_file])

        result = subprocess.call(cmd)
        if result != 0:
            raise VendorizeError("Failed to vendor {!r}".format(vendor_spec))

        # Temporarily move importable code into the vendor chroot for importing tools later.
        site_packages = find_site_packages(prefix_dir=prefix_dir_by_vendor_spec[vendor_spec])
        safe_mkdir(vendor_spec.target_dir)
        for item in os.listdir(site_packages):
            src = os.path.join(site_packages, item)
            dst = os.path.join(vendor_spec.target_dir, item)
            shutil.move(src, dst)
            moved[dst] = src

        if vendor_spec.constrain:
            cmd = ["pip", "freeze", "--all", "--path", vendor_spec.target_dir]
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
            stdout, _ = process.communicate()
            if process.returncode != 0:
                raise VendorizeError("Failed to freeze vendoring of {!r}".format(vendor_spec))
            with open(constraints_file, "wb") as fp:
                fp.write(stdout)

        vendor_spec.create_packages()

    vendored_path = [vendor_spec.target_dir for vendor_spec in vendor_specs if vendor_spec.rewrite]
    import_rewriter = ImportRewriter.for_path_items(prefix=prefix, path_items=vendored_path)

    rewrite_paths = [os.path.join(root_dir, c) for c in ("pex", "tests")] + vendored_path
    for rewrite_path in rewrite_paths:
        for root, dirs, files in os.walk(rewrite_path):
            if root == os.path.join(root_dir, "pex", "vendor"):
                dirs[:] = [d for d in dirs if d != "_vendored"]
            for f in files:
                if f.endswith(".py"):
                    python_file = os.path.join(root, f)
                    print(green("Examining {python_file}...".format(python_file=python_file)))
                    modifications = import_rewriter.rewrite(python_file)
                    if modifications:
                        num_mods = len(modifications)
                        print(
                            bold(
                                green(
                                    "  Vendorized {count} import{plural} in {python_file}".format(
                                        count=num_mods,
                                        plural="s" if num_mods > 1 else "",
                                        python_file=python_file,
                                    )
                                )
                            )
                        )
                        for _from, _to in modifications.items():
                            print("    {} -> {}".format(_from, _to))

    # Import all code needed below now before we move any vendored bits it depends on temporarily
    # back to the prefix site-packages dir.
    from pex.dist_metadata import find_distributions
    from pex.pep_376 import Record

    dist_by_vendor_spec = OrderedDict()
    for vendor_spec in vendor_specs:
        for dist in find_distributions(search_path=[vendor_spec.target_dir]):
            if dist.project_name == vendor_spec.key:
                dist_by_vendor_spec[vendor_spec] = dist
                break
        if vendor_spec not in dist_by_vendor_spec:
            raise RuntimeError("Failed to find a distribution for {}.".format(vendor_spec))

    # Move the importables back to their temporary dir chroot original locations.
    for dst, src in moved.items():
        shutil.move(dst, src)

    for vendor_spec in vendor_specs:
        print(
            bold(
                green(
                    "Fixing up scripts and distribution metadata for {requirement}".format(
                        requirement=vendor_spec.requirement
                    )
                )
            )
        )
        # Move the `pip install --prefix` temporary dir chroot into the vendor chroot.
        prefix_dir = prefix_dir_by_vendor_spec[vendor_spec]
        for item in os.listdir(prefix_dir):
            shutil.move(os.path.join(prefix_dir, item), os.path.join(vendor_spec.target_dir, item))
        dist = dist_by_vendor_spec[vendor_spec]

        # Finally, let Record fixup the chroot to its final importable form.
        record = Record.from_prefix_install(
            prefix_dir=vendor_spec.target_dir,
            project_name=dist.project_name,
            version=dist.version,
        )
        record.fixup_install(exclude=("constraints.txt", "__init__.py", "__pycache__"))


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--no-update",
        dest="update",
        default=True,
        action="store_false",
        help="Do not update vendored project versions, just vendorize the existing versions afresh.",
    )
    options = parser.parse_args()

    root_directory = VendorSpec.ROOT
    try:
        safe_rmtree(VendorSpec.vendor_root())
        vendorize(
            root_dir=root_directory,
            vendor_specs=list(iter_vendor_specs()),
            prefix="pex.third_party",
            update=options.update,
        )
        sys.exit(0)
    except VendorizeError as e:
        print("Problem encountered vendorizing: {}".format(e), file=sys.stderr)
        sys.exit(1)



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess
from textwrap import dedent

import pytest

from pex.dist_metadata import Distribution
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pex import PEX
from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Any, Iterable, Mapping


@pytest.fixture(scope="module")
def td(tmpdir_factory):
    # type: (Any) -> Any
    return tmpdir_factory.mktemp("td")


PIP_PROJECT_NAME = ProjectName("pip")
SETUPTOOLS_PROJECT_NAME = ProjectName("setuptools")


def index_distributions(dists):
    # type: (Iterable[Distribution]) -> Mapping[ProjectName, Version]
    return {dist.metadata.project_name: dist.metadata.version for dist in dists}


@pytest.fixture(scope="module")
def baseline_venv_with_pip(td):
    # type: (Any) -> Mapping[ProjectName, Version]
    baseline_venv = Virtualenv.create(venv_dir=str(td.join("baseline.venv")))
    baseline_venv.install_pip()
    baseline_venv_distributions = index_distributions(baseline_venv.iter_distributions())
    assert {PIP_PROJECT_NAME, SETUPTOOLS_PROJECT_NAME} == set(baseline_venv_distributions)
    return baseline_venv_distributions


@pytest.fixture(scope="module")
def baseline_venv_pip_version(baseline_venv_with_pip):
    # type: (Mapping[ProjectName, Version]) -> Version
    return baseline_venv_with_pip[PIP_PROJECT_NAME]


@pytest.fixture(scope="module")
def baseline_venv_setuptools_version(baseline_venv_with_pip):
    # type: (Mapping[ProjectName, Version]) -> Version
    return baseline_venv_with_pip[SETUPTOOLS_PROJECT_NAME]


def assert_venv_dists(
    venv_dir,  # type: str
    expected_pip_version,  # type: Version
    expected_setuptools_version,  # type: Version
):
    virtualenv = Virtualenv(venv_dir)
    dists = index_distributions(virtualenv.iter_distributions())
    assert expected_pip_version == dists[PIP_PROJECT_NAME]
    assert expected_setuptools_version == dists[SETUPTOOLS_PROJECT_NAME]

    def reported_version(module):
        # type: (str) -> Version
        return Version(
            subprocess.check_output(
                args=[
                    virtualenv.interpreter.binary,
                    "-c",
                    "import {module}; print({module}.__version__)".format(module=module),
                ]
            ).decode("utf-8")
        )

    assert expected_pip_version == reported_version("pip")
    assert expected_setuptools_version == reported_version("setuptools")


def assert_venv_dists_no_conflicts(
    tmpdir,  # type: Any
    pex,  # type: str
    expected_pip_version,  # type: Version
    expected_setuptools_version,  # type: Version
):
    # type: (...) -> None
    venv_dir = os.path.join(str(tmpdir), "venv_dir")
    subprocess.check_call(args=[pex, "venv", "--pip", venv_dir], env=make_env(PEX_TOOLS=1))
    assert_venv_dists(venv_dir, expected_pip_version, expected_setuptools_version)


def test_pip_empty_pex(
    tmpdir,  # type: Any
    baseline_venv_pip_version,  # type: Version
    baseline_venv_setuptools_version,  # type: Version
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["-o", pex, "--include-tools"]).assert_success()

    assert_venv_dists_no_conflicts(
        tmpdir,
        pex,
        expected_pip_version=baseline_venv_pip_version,
        expected_setuptools_version=baseline_venv_setuptools_version,
    )


def test_pip_pex_no_conflicts(
    tmpdir,  # type: Any
    baseline_venv_pip_version,  # type: Version
    baseline_venv_setuptools_version,  # type: Version
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[
            "-o",
            pex,
            "pip=={version}".format(version=baseline_venv_pip_version),
            "setuptools=={version}".format(version=baseline_venv_setuptools_version),
            "--include-tools",
        ]
    ).assert_success()

    assert_venv_dists_no_conflicts(
        tmpdir,
        pex,
        expected_pip_version=baseline_venv_pip_version,
        expected_setuptools_version=baseline_venv_setuptools_version,
    )


def assert_venv_dists_conflicts(
    tmpdir,  # type: Any
    pex,  # type: str
    baseline_venv_pip_version,  # type: Version
    baseline_venv_setuptools_version,  # type: Version
    expected_pip_version,  # type: Version
    expected_setuptools_version,  # type: Version
):
    # type: (...) -> None

    expected_conflicts = []
    if baseline_venv_pip_version != expected_pip_version:
        expected_conflicts.append("pip {version}".format(version=expected_pip_version))
    if baseline_venv_setuptools_version != expected_setuptools_version:
        expected_conflicts.append(
            "setuptools {version}".format(version=expected_setuptools_version)
        )
    assert (
        expected_conflicts
    ), "The assert_venv_dists_conflicts function requires at least one conflict."

    venv_dir = os.path.join(str(tmpdir), "venv_dir")
    args = [pex, "venv", "--pip", venv_dir]

    expected_message_prefix = (
        dedent(
            """\
            You asked for --pip to be installed in the venv at {venv_dir},
            but the PEX at {pex} already contains:
            {conflicts}
            """
        )
        .format(venv_dir=venv_dir, pex=pex, conflicts=os.linesep.join(expected_conflicts))
        .strip()
    )

    process = subprocess.Popen(args, stderr=subprocess.PIPE, env=make_env(PEX_TOOLS=1))
    _, stderr = process.communicate()
    assert 0 != process.returncode

    decoded_stderr = stderr.decode("utf-8")
    assert (
        dedent(
            """\
            {prefix}
            Consider re-running either without --pip or with --collisions-ok.
            """
        ).format(prefix=expected_message_prefix)
        in decoded_stderr
    ), decoded_stderr

    process = subprocess.Popen(
        args + ["--force", "--collisions-ok"], stderr=subprocess.PIPE, env=make_env(PEX_TOOLS=1)
    )
    _, stderr = process.communicate()
    assert 0 == process.returncode
    decoded_stderr = stderr.decode("utf-8")
    assert (
        dedent(
            """\
            {prefix}
            Uninstalling venv versions and using versions from the PEX.
            """
        ).format(prefix=expected_message_prefix)
        in decoded_stderr
    ), decoded_stderr

    assert_venv_dists(venv_dir, expected_pip_version, expected_setuptools_version)


def test_pip_pex_pip_conflict(
    tmpdir,  # type: Any
    baseline_venv_pip_version,  # type: Version
    baseline_venv_setuptools_version,  # type: Version
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[
            "-o",
            pex,
            "pip!={version}".format(version=baseline_venv_pip_version),
            "--include-tools",
        ]
    ).assert_success()
    pex_pip_version = index_distributions(PEX(pex).resolve())[PIP_PROJECT_NAME]

    assert_venv_dists_conflicts(
        tmpdir,
        pex,
        baseline_venv_pip_version=baseline_venv_pip_version,
        baseline_venv_setuptools_version=baseline_venv_setuptools_version,
        expected_pip_version=pex_pip_version,
        expected_setuptools_version=baseline_venv_setuptools_version,
    )


def test_pip_pex_setuptools_conflict(
    tmpdir,  # type: Any
    baseline_venv_pip_version,  # type: Version
    baseline_venv_setuptools_version,  # type: Version
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[
            "-o",
            pex,
            "setuptools!={version}".format(version=baseline_venv_setuptools_version),
            "--include-tools",
        ]
    ).assert_success()
    pex_setuptools_version = index_distributions(PEX(pex).resolve())[SETUPTOOLS_PROJECT_NAME]

    assert_venv_dists_conflicts(
        tmpdir,
        pex,
        baseline_venv_pip_version=baseline_venv_pip_version,
        baseline_venv_setuptools_version=baseline_venv_setuptools_version,
        expected_pip_version=baseline_venv_pip_version,
        expected_setuptools_version=pex_setuptools_version,
    )


def test_pip_pex_both_conflict(
    tmpdir,  # type: Any
    baseline_venv_pip_version,  # type: Version
    baseline_venv_setuptools_version,  # type: Version
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[
            "-o",
            pex,
            "pip!={version}".format(version=baseline_venv_pip_version),
            "setuptools!={version}".format(version=baseline_venv_setuptools_version),
            "--include-tools",
        ]
    ).assert_success()
    pex_pip_version = index_distributions(PEX(pex).resolve())[PIP_PROJECT_NAME]
    pex_setuptools_version = index_distributions(PEX(pex).resolve())[SETUPTOOLS_PROJECT_NAME]

    assert_venv_dists_conflicts(
        tmpdir,
        pex,
        baseline_venv_pip_version=baseline_venv_pip_version,
        baseline_venv_setuptools_version=baseline_venv_setuptools_version,
        expected_pip_version=pex_pip_version,
        expected_setuptools_version=pex_setuptools_version,
    )



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import IO, Dict, List, Mapping, Optional, Tuple

    Value = Optional[str]
    Attributes = Mapping[str, Value]


class DiGraph(object):
    """Renders a dot digraph built up from nodes and edges."""

    @staticmethod
    def _render_ID(value):
        # type: (str) -> str
        # See https://graphviz.org/doc/info/lang.html for the various forms of `ID`.
        return '"{}"'.format(value.replace('"', '\\"'))

    @classmethod
    def _render_a_list(cls, attributes):
        # type: (Attributes) -> str
        # See https://graphviz.org/doc/info/lang.html for the `a_list` production.
        return ", ".join(
            "{name}={value}".format(name=name, value=cls._render_ID(value))
            for name, value in attributes.items()
            if value is not None
        )

    def __init__(
        self,
        name,  # type: str
        strict=True,  # type: bool
        **attributes  # type: Value
    ):
        # type: (...) -> None
        """
        :param name: A name for the graph.
        :param strict: Whether or not duplicate edges are collapsed into one edge.
        """
        self._name = name
        self._strict = strict
        self._attributes = attributes  # type: Attributes
        self._nodes = {}  # type: Dict[str, Attributes]
        self._edges = []  # type: List[Tuple[str, str, Attributes]]

    @property
    def name(self):
        return self._name

    def add_node(
        self,
        name,  # type: str
        **attributes  # type: Value
    ):
        # type: (...) -> None
        """Adds a node to the graph.

        This is done implicitly by add_edge for the nodes the edge connects, but may be useful when
        the node is either isolated or else needs to be decorated with attributes.

        :param name: The name of the node.
        """
        self._nodes[name] = attributes

    def add_edge(
        self,
        start,  # type: str
        end,  # type: str
        **attributes  # type: Value
    ):
        # type: (...) -> None
        """

        :param start: The name of the start node.
        :param end: The name of the end node.
        :param attributes: Any extra attributes for the edge connecting the start node to the end
                           node.
        """
        self._edges.append((start, end, attributes))

    def emit(self, out):
        # type: (IO[str]) -> None
        """Render the current state of this digraph to the given `out` stream.

        :param out: A stream to render this digraph to. N/B.: Will not be flushed or closed.
        """

        def emit_attr_stmt(
            stmt,  # type: str
            attributes,  # type: Attributes
        ):
            # type: (...) -> None
            # See https://graphviz.org/doc/info/lang.html for the `attr_stmt` production.
            out.write(
                "{statement} [{a_list}];\n".format(
                    statement=stmt, a_list=self._render_a_list(attributes)
                )
            )

        if self._strict:
            out.write("strict ")
        out.write("digraph {name} {{\n".format(name=self._render_ID(self._name)))
        emit_attr_stmt("graph", self._attributes)
        for node, attributes in self._nodes.items():
            emit_attr_stmt(self._render_ID(node), attributes)
        for start, end, attributes in self._edges:
            emit_attr_stmt(
                "{start} -> {end}".format(start=self._render_ID(start), end=self._render_ID(end)),
                attributes,
            )
        out.write("}\n")
